# César Vonc - code.vonc.fr - v 1.2 - décembre 2013

import os
import sys
import math
import hashlib
import c4d
from c4d import plugins, utils, bitmaps, gui, Vector, Matrix
from c4d.utils import HPBToMatrix, MatrixScale

MODULE_ID = 1029973
VONC_DEFPOINT_NOM = 1000
VONC_DEFPOINT_RAYON = 1000
VONC_DEFPOINT_INTENSITE = 1001
VONC_DEFPOINT_POSITION = 1002
VONC_DEFPOINT_ATTENUATION = 1003
VONC_DEFPOINT_ECHELLE = 1004
VONC_DEFPOINT_ROTATION = 1006

class DeformateurPoint(plugins.ObjectData) :
	PRISES = 3
	opt_cle = ""
	opt_cle_ans = None
	
	def Message(self, node, type, data) :
		if type == c4d.MSG_MENUPREPARE :
			node.SetDeformMode(True)
		return True
	
	def Init(self, op) :
		donnees = op.GetDataInstance()
		courbe = c4d.SplineData()
		courbe.MakeLinearSplineBezier(2)
		courbe.SetKnot(index = 0, vPos = Vector(0.0), vTangentLeft = Vector(-0.4, 0.0, 0.0), vTangentRight = Vector(0.4, 0.0, 0.0))
		courbe.SetKnot(index = 1, vPos = Vector(1.0, 1.0, 0.0), vTangentLeft = Vector(-0.4, 0.0, 0.0), vTangentRight = Vector(0.4, 0.0, 0.0))
		
		self.InitAttr(op, float, [VONC_DEFPOINT_RAYON])
		self.InitAttr(op, float, [VONC_DEFPOINT_INTENSITE])
		self.InitAttr(op, Vector, [VONC_DEFPOINT_POSITION])
		self.InitAttr(op, c4d.SplineData, [VONC_DEFPOINT_ATTENUATION])
		self.InitAttr(op, Vector, [VONC_DEFPOINT_ECHELLE])
		self.InitAttr(op, Vector, [VONC_DEFPOINT_ROTATION])
		
		op[VONC_DEFPOINT_RAYON] = 200.0
		op[VONC_DEFPOINT_INTENSITE] = 1.0
		op[VONC_DEFPOINT_POSITION] = Vector(0.0, 100.0, 0.0)
		op[VONC_DEFPOINT_ATTENUATION] = courbe
		op[VONC_DEFPOINT_ECHELLE] = Vector(1.0)
		op[VONC_DEFPOINT_ROTATION] = Vector()
		
		return True
	
	def GetDimension(self, op, mp, rad) :
		mp.x = 0.0
		mp.y = 0.0
		mp.z = 0.0
		
		rayon = op[VONC_DEFPOINT_RAYON]
		rad.x = rayon
		rad.y = rayon
		rad.z = rayon
	
	def GetHandleCount(op) :
		return DeformateurPoint.PRISES

	def GetHandle(self, op, i, info) :
		donnees = op.GetDataInstance()
		if donnees is None : return
		
		if i == 0 :
			info.position = Vector(donnees.GetReal(VONC_DEFPOINT_RAYON), 0.0, 0.0)
			info.direction = Vector(1.0, 0.0, 0.0)
			info.type = c4d.HANDLECONSTRAINTTYPE_LINEAR
		
		elif i == 1 :
			info.position = Vector(donnees.GetReal(VONC_DEFPOINT_INTENSITE) * 250.0, 0.0, 0.0)
			info.direction = Vector(1.0, 0.0, 0.0)
			info.type = c4d.HANDLECONSTRAINTTYPE_LINEAR
			
		elif i == 2 :
			info.position = donnees.GetVector(VONC_DEFPOINT_POSITION)
			info.direction = donnees.GetVector(VONC_DEFPOINT_POSITION)
			info.type = c4d.HANDLECONSTRAINTTYPE_LINEAR
	
	def DetectHandle(self, op, bd, x, y, qualifier) :
		if qualifier & c4d.QUALIFIER_CTRL : return c4d.NOTOK
		
		mg = op.GetMg()
		ret = c4d.NOTOK
		
		for i in xrange(self.GetHandleCount()) :
			info = c4d.HandleInfo()
			self.GetHandle(op, i, info)
			if bd.PointInRange(info.position * mg, x, y) :
				ret = i
				if not qualifier & c4d.QUALIFIER_SHIFT : break
		
		return ret
	

	def MoveHandle(self, op, undo, mouse_pos, hit_id, qualifier, bd) :
		dst = op.GetDataInstance()
		
		info = c4d.HandleInfo()
		val = mouse_pos.x
		self.GetHandle(op, hit_id, info)
		
		if bd is not None :
			mg = op.GetUpMg() * undo.GetMl()
			pos = bd.ProjectPointOnLine(info.position * mg, info.direction ^ mg, mouse_pos.x, mouse_pos.y)
			val = (pos[0] * ~mg) * info.direction
		
		if hit_id == 0 :
			dst.SetReal(VONC_DEFPOINT_RAYON, utils.FCut(val, 0.0, sys.maxint))
		
		elif hit_id == 1 :
			dst.SetReal(VONC_DEFPOINT_INTENSITE, utils.FCut(val * 0.004, 0.0, 1.0))
			
		elif hit_id == 2 :
			rot = bd.GetMg().off - mg.off
			pos = bd.ProjectPointOnPlane(mg.off, rot, mouse_pos.x, mouse_pos.y)
			dst.SetVector(VONC_DEFPOINT_POSITION, pos[0] * ~mg)
		
		return True


	def Draw(self, op, drawpass, bd, bh) :
		if drawpass == c4d.DRAWPASS_OBJECT :
			rayon = op[VONC_DEFPOINT_RAYON]
			m = bh.GetMg()
			
			m.v1 *= rayon
			m.v2 *= rayon
			m.v3 *= rayon
			
			bd.SetMatrix_Matrix(None, c4d.Matrix())
			bd.SetPen(bd.GetObjectColor(bh, op))
			bd.DrawCircle(m)
			h = m.v2
			m.v2 = m.v3
			m.v3 = h
			bd.DrawCircle(m)
			h = m.v1
			m.v1 = m.v3
			m.v3 = h
			bd.DrawCircle(m)
			
			bd.DrawLine(bh.GetMg().Mul(op[VONC_DEFPOINT_POSITION]), bh.GetMg().off, 0)
			
			return c4d.DRAWRESULT_OK
		
		elif drawpass == c4d.DRAWPASS_HANDLES :
			i = 0
			bd.SetMatrix_Matrix(None, bh.GetMg())
			
			surbrille = op.GetHighlightHandle(bd)
			
			for i in xrange(self.PRISES) :
				info = c4d.HandleInfo()
				self.GetHandle(op, i, info)
				if i == surbrille :
					bd.SetPen(c4d.GetViewColor(c4d.VIEWCOLOR_SELECTION_PREVIEW))
				else :
					bd.SetPen(c4d.GetViewColor(c4d.VIEWCOLOR_ACTIVEPOINT))
					
				bd.DrawHandle(info.position, c4d.DRAWHANDLE_BIG, 0)
			
			info = c4d.HandleInfo()
			self.GetHandle(op, 1, info)
			bd.SetPen(c4d.GetViewColor(c4d.VIEWCOLOR_ACTIVEPOINT))
			bd.DrawLine(info.position, Vector(0), 0)
			
			return c4d.DRAWRESULT_OK
		
		return c4d.DRAWRESULT_SKIP
	
	
	def ModifyObject(self, mod, doc, op, op_mg, mod_mg, lod, flags, thread) :
		rayon = mod[VONC_DEFPOINT_RAYON]
		intensite = mod[VONC_DEFPOINT_INTENSITE]
		position = mod[VONC_DEFPOINT_POSITION]
		courbe = mod[VONC_DEFPOINT_ATTENUATION]
		echelle = mod[VONC_DEFPOINT_ECHELLE]
		rotation = mod[VONC_DEFPOINT_ROTATION]
		
		if not op.CheckType(c4d.Opoint) : return True
		
		points = op.GetAllPoints()
		nbPts = op.GetPointCount()
		
		if not nbPts : return True
		
		m = mod_mg.__rmul__(~op_mg) # Matrice de mod relative à op
		position = mod_mg.MulV(position)
		
		matrice = HPBToMatrix(rotation)
		matrice.Scale(echelle)
		
		weight = op.CalcVertexmap(mod)
		
		if echelle == Vector(1.0) : echelle = None
		if rotation == Vector() : rotation = None
		if position == Vector() : position = None
		
		dec = mod_mg.off
		
		for i, p in enumerate(points) :
			s = intensite
			if not s or not rayon : break
			
			if weight is not None : s *= weigth[i]
			
			fac = rayon - (p - m.off).GetLength()
			if fac < 0 : fac = 0
			else : fac = courbe.GetPoint(fac / rayon).y
			
			f = fac * s
			p = op_mg.Mul(p)
			
			if (rotation or echelle) : p = (matrice.MulV(p - dec) + dec) * f  + p * (1.0 - f)
			if (position) : p = p + position * f
			
			p = (~op_mg).Mul(p)
			
			op.SetPoint(i, p)
		
		op.Message(c4d.MSG_UPDATE)
		
		return True

        

if __name__ == "__main__":
	path, fn = os.path.split(__file__)
	bmp = bitmaps.BaseBitmap()
	bmp.InitWith(os.path.join(path, "res", "vonc_defpoint.tif"))
	plugins.RegisterObjectPlugin(id=MODULE_ID, str=c4d.plugins.GeLoadString(VONC_DEFPOINT_NOM)+' v1.2',
								g=DeformateurPoint,
								description="vonc_defpoint", icon=bmp,
								info=c4d.OBJECT_MODIFIER)
